///////////////////////////////////////////////////////////////////////////////
//
// Filename: CsvLoader.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  Comma Separated Variable file loader.
//
// (c) Copyright R.H.Thornburrow, 2001
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.csv;

import com.tsac.familytree.*;
import com.tsac.familytree.gui.*;

import java.util.*;
import java.io.IOException;
import java.io.FileReader;
import java.io.BufferedReader;

public class CsvLoader
    implements FamilyTreeLoader
{
    private Person[] _people;
    private Marriage[] _marriages;

	public CsvLoader()
	{
	}
	
	public Person[] getPeople()  { return _people; }
	public Marriage[] getMarriages()  { return _marriages; }
	public Person[] getInvertedRoots()  { return null; }	
	public Person[] getConventionalRoots()  { return null; }
	
	public void load( String filename )
    {
        Vector people = new Vector();
        Vector marriages = new Vector();
        Hashtable keys = new Hashtable();

        try
        {
            FileReader frd = new FileReader( filename );
            BufferedReader istrm = new BufferedReader( frd );
            
            String input;
            int mode = 0;
            int line = 0;
            
            while( (mode != -1) && (input = istrm.readLine()) != null )
            {
                line++;
                
                if( (input.length() == 0) || (input.charAt(0) == '#') )
                    continue;
                
                if( input.charAt(0) == '@' )
                {
                    mode++;
                    continue;
                }
                
                try
                {
                    switch( mode )
                    {
                    case 1: // person
                        parsePerson( input, people, keys );
                        break;
                    case 2: // marriage
                        parseMarriage( input, marriages, keys );
                        break;
                    default:
                        System.out.println( "PARSE ERROR: " + line + ": Unexpected input (" + input + ")" );
                        mode = -1;
                        break;
                    }
                }
                catch( NoSuchElementException e )
                {
                    System.out.println( "PARSE ERROR: " + line + ": Not enough commas (" + input + ")" );
                }
                catch( ParseException e )
                {
                    System.out.println( "PARSE ERROR: " + line + ": " + e.getMessage() + " (" + input + ")" );
                }
            }
            
            frd.close();
            
            System.out.println( "Loaded " + people.size() + " people and " + marriages.size() + " marriages" );
            
            _people = new Person[ people.size() ];
            people.copyInto( _people );
            
            _marriages = new Marriage[ marriages.size() ];
            marriages.copyInto( _marriages );
        }
        catch( IOException e )
        {
            e.printStackTrace();
        }
    }
    
    private void parsePerson( String input, Vector people, Hashtable keys )
        throws ParseException
    {
        Person mother = null;
        Person father = null;
        Person node = null;
        
        StringTokenizer tok = new StringTokenizer( input, ",", true );
        
        String key = getToken( tok, true );
        if( key == null ) throw new ParseException( "Missing key" );
        if( key.length() != 7 ) throw new ParseException( "Key must be 7 chars" );
        
        String sex = getToken( tok, true );
        if( sex == null ) throw new ParseException( "Missing sex" );
        
        String surname = getToken( tok, true );
        if( surname == null ) throw new ParseException( "Missing surname" );
        
        String forenames = getToken( tok, true );
        if( forenames == null ) throw new ParseException( "Missing forenames" );
        
        String born = getToken( tok, true );
        String christened = getToken( tok, true );
        String died = getToken( tok, true );
        String motherkey = getToken( tok, true );
        
        String fatherkey = null;
        if( tok.hasMoreTokens() )
            fatherkey = getToken( tok, false );
            
        if( (fatherkey != null) && tok.hasMoreTokens() && !",".equals( tok.nextToken() ) )
            throw new ParseException( "Bad person at father" );
        
        String page = null;
        if( tok.hasMoreTokens() )
            page = getToken( tok, false );

        if( tok.hasMoreTokens() )
            throw new ParseException( "Extraneous tokens" );
        
        if( motherkey != null )
        {
            if( motherkey.length() != 7 ) throw new ParseException( "Mother key must be 7 chars" );
            mother = (Person) keys.get( motherkey );
            if( mother == null  ) throw new ParseException( "Mother doesn't exist" );
            if( !(mother instanceof Female) ) throw new ParseException( "Mother must be female" );
        }
            
        if( fatherkey != null )
        {
            if( fatherkey.length() != 7 ) throw new ParseException( "Father key must be 7 chars" );
            father = (Person) keys.get( fatherkey );
            if( father == null  ) throw new ParseException( "Father doesn't exist" );
            if( !(father instanceof Male) ) throw new ParseException( "Father must be male" );
        }
        
        if( "M".equals( sex ) )
            node = new Male( key, surname, forenames, born, christened, died, mother, father, page );
        else if( "F".equals( sex ) )
            node = new Female( key, surname, forenames, born, christened, died, mother, father, page );
        else
            throw new ParseException( "Sex must be M or F" );
        
        people.addElement( node );
        keys.put( key, node );
    }

    private void parseMarriage( String input, Vector marriages, Hashtable keys )
        throws ParseException
    {
        Person husband = null;
        Person wife = null;
        
        StringTokenizer tok = new StringTokenizer( input, ",", true );
        
        String husbandkey = getToken( tok, true );
        if( husbandkey == null ) throw new ParseException( "Missing Husband key" );

        String wifekey = getToken( tok, true );
        if( wifekey == null ) throw new ParseException( "Missing Wife key" );

        String date = null;
        if( tok.hasMoreTokens() )
            date = getToken( tok, false );
        
        if( husbandkey.length() != 7 ) throw new ParseException( "Husband key must be 7 chars" );
        husband = (Person) keys.get( husbandkey );
        if( husband == null  ) throw new ParseException( "Husband doesn't exist" );
        if( !(husband instanceof Male) ) throw new ParseException( "Husband must be male" );
        
        if( wifekey.length() != 7 ) throw new ParseException( "Wife key must be 7 chars" );
        wife = (Person) keys.get( wifekey );
        if( wife == null  ) throw new ParseException( "Wife doesn't exist" );
        if( !(wife instanceof Female) ) throw new ParseException( "Wife must be male" );
        
        marriages.addElement( new Marriage( husband, wife, date ) );
    }
    
    private String getToken( StringTokenizer tok, boolean comma )
        throws ParseException
    {
        String str = tok.nextToken();
        if( ",".equals( str ) ) return null;
        
        if( comma )
        {
            String str2 = tok.nextToken();
            if( !",".equals( str2 ) ) throw new ParseException( "Missing comma" );
        }
        
        return str;
    }
}