///////////////////////////////////////////////////////////////////////////////
//
// Filename: FtmlHandler.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  FTML SAX Handler for use with Xerces.
//
// (c) Copyright R.H.Thornburrow, 2001
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.ftml.parser;

import com.tsac.familytree.*;

import org.xml.sax.SAXException;
import org.xml.sax.Attributes;

import java.util.Hashtable;
import java.util.Vector;

public class FtmlHandler
    extends org.xml.sax.helpers.DefaultHandler
{
    private Hashtable _elements = new Hashtable();
    private Vector _people = new Vector();
    private Vector _roots = new Vector();
    private Vector _marriages = new Vector();
    
    private Person _currPerson = null;
    private String _currId = null;
    
    public static final int FTML      = 1;
    public static final int PEOPLE    = 2;
    public static final int MARRIAGES = 3;
    public static final int PERSON    = 4;
    public static final int MARRIAGE  = 5;
    public static final int BORN      = 6;
    public static final int CHRISTEN  = 7;
    public static final int DIED      = 8;
    public static final int MOTHER    = 9;
    public static final int FATHER    = 10;
    public static final int PAGE      = 11;
    public static final int CENSUS    = 12;
    
    
    /**
     * Constructor.
     */
    public FtmlHandler()
    {
        _elements.put( "ftml",      new Integer( FTML ) );
        _elements.put( "people",    new Integer( PEOPLE ) );
        _elements.put( "marriages", new Integer( MARRIAGES ) );
        _elements.put( "person",    new Integer( PERSON ) );
        _elements.put( "marriage",  new Integer( MARRIAGE ) );
        _elements.put( "born",      new Integer( BORN ) );
        _elements.put( "christened", new Integer( CHRISTEN ) );
        _elements.put( "died",      new Integer( DIED ) );
        _elements.put( "mother",    new Integer( MOTHER ) );
        _elements.put( "father",    new Integer( FATHER ) );
        _elements.put( "page",      new Integer( PAGE ) );
        _elements.put( "census",    new Integer( CENSUS ) );
    }


    public Vector getPeople()  { return _people; }
    public Vector getRoots()  { return _roots; }
    public Vector getMarriages()  { return _marriages; }


    /**
     * Handles the start of each element.  Creates a new object for those that
     * need one, or sets an property for those that don't.
     */
    public void startElement( String uri, String localName, String qName, Attributes attributes )
        throws SAXException
    {
        switch( lookupElement(qName) )
        {
        case PERSON:
            if( _currPerson != null )
                throw new SAXException( "SEMANTIC ERROR: <person> with Person already open" );
            _currPerson = createPerson( attributes );
            break;
        case BORN:
            applyBorn( _currPerson, attributes );
            break;
        case CHRISTEN:
            applyChristen( _currPerson, attributes );
            break;
        case DIED:
            applyDied( _currPerson, attributes );
            break;
        case MOTHER:
            applyMother( _currPerson, attributes );
            break;
        case FATHER:
            applyFather( _currPerson, attributes );
            break;
        case PAGE:
            applyPage( _currPerson, attributes );
            break;
        case MARRIAGE:
            createMarriage( attributes );
            break;
        }
    }
    
    
    /**
     * Handles the end of each element.  Saves any objects that need saving.
     */
    public void endElement( String uri, String localName, String qName )
        throws SAXException
    {
        switch( lookupElement(qName) )
        {
        case PERSON:
            if( (_currPerson == null) || (_currId == null) )
                throw new SAXException( "SEMANTIC ERROR: </person> with Person already closed" );
            _people.addElement( _currPerson );
            _currId = null;
            _currPerson = null;
            break;
        }
    }
    
    
    /**
     * Checks that an element is part of the FTML vocabulary and turns it into
     * a token that can be used in a switch statement.  Throws an exception if
     * the element cannot be found.
     */
    private int lookupElement( String qName )
        throws SAXException
    {
        Integer element = (Integer) _elements.get( qName );
        
        if( element == null )
            throw new SAXException( "PARSE ERROR: unrecognised element " + qName );
        
        return element.intValue();
    }
    
    
    private Person createPerson( Attributes attrs )
    {
        Person ret;
        
        _currId = attrs.getValue( "id" );
        
        String sex       = attrs.getValue( "sex" );
        String surname   = attrs.getValue( "surname" );
        String forenames = attrs.getValue( "forenames" );
        String root      = attrs.getValue( "root" );
        
        if( "male".equals( sex ) )
            ret = new Male( _currId, surname, forenames );
        else    
            ret = new Female( _currId, surname, forenames );
        
        if( (root != null) && root.equals( "true" ) )
            _roots.addElement( ret );
        
        return ret;
    }
    
    
    private void applyBorn( Person person, Attributes attrs )
        throws SAXException
    {
        if( person == null )
            throw new SAXException( "SEMANTIC ERROR: <born /> with Person already closed" );
            
        String date   = attrs.getValue( "date" );
        String place  = attrs.getValue( "place" );
        String href   = attrs.getValue( "href" );
        
        person.setBorn( date );
        if( place != null ) person.setBornPlace( place );
        if( href != null )  person.setBornPage( href );
    }
    
    
    private void applyChristen( Person person, Attributes attrs )
        throws SAXException
    {
        if( person == null )
            throw new SAXException( "SEMANTIC ERROR: <christened /> with Person already closed" );
            
        String date   = attrs.getValue( "date" );
        String place  = attrs.getValue( "place" );
        String href   = attrs.getValue( "href" );
        
        person.setChristened( date );
        if( place != null ) person.setChristenPlace( place );
        if( href != null )  person.setChristenPage( href );
    }
    
    
    private void applyDied( Person person, Attributes attrs )
        throws SAXException
    {
        if( person == null )
            throw new SAXException( "SEMANTIC ERROR: <died /> with Person already closed" );
            
        String date   = attrs.getValue( "date" );
        String place  = attrs.getValue( "place" );
        String href   = attrs.getValue( "href" );
        
        person.setDied( date );
        if( place != null ) person.setDiedPlace( place );
        if( href != null )  person.setDiedPage( href );
    }
    
    
    private void applyMother( Person person, Attributes attrs )
        throws SAXException
    {
        if( person == null )
            throw new SAXException( "SEMANTIC ERROR: <mother /> with Person already closed" );
            
        String id   = attrs.getValue( "id" );

        person.setMother( id );
    }
    
    
    private void applyFather( Person person, Attributes attrs )
        throws SAXException
    {
        if( person == null )
            throw new SAXException( "SEMANTIC ERROR: <father /> with Person already closed" );
            
        String id   = attrs.getValue( "id" );
        
        person.setFather( id );
    }
    
    
    private void applyPage( Person person, Attributes attrs )
        throws SAXException
    {
        if( person == null )
            throw new SAXException( "SEMANTIC ERROR: <page /> with Person already closed" );
            
        String href   = attrs.getValue( "href" );
        
        person.setPage( href );
    }
    
    
    private void createMarriage( Attributes attrs )
        throws SAXException
    {
        String husbandid = attrs.getValue( "husband" );
        String wifeid    = attrs.getValue( "wife" );
        String date      = attrs.getValue( "date" );
        String place     = attrs.getValue( "place" );   // NOT USED
        String href      = attrs.getValue( "href" );
        
        Person husband = Person.getPerson( husbandid );
        
        if( husband == null )
            throw new SAXException( "SEMANTIC ERROR: Person with id='" + husbandid + "' does not exist" );

        husband.setMarriedPage( href );

        Person wife    = Person.getPerson( wifeid );

        if( wife == null )
            throw new SAXException( "SEMANTIC ERROR: Person with id='" + wifeid + "' does not exist" );

        wife.setMarriedPage( href );
        
        _marriages.addElement( new Marriage( husband, wife, date ) );
    }
}